from __future__ import annotations
from typing import TYPE_CHECKING, Callable, Dict, List, Any, Optional, Tuple, TypeVar, cast
from functools import wraps
from random import shuffle
import abc
from .types import (
    DataRow,
    Params,
    Data,
    Record,
)
if TYPE_CHECKING:
    from .table import SqliteTable


class _Str(abc.ABC):

    @abc.abstractmethod
    def _str(self, params: Params) -> str:
        ...


class Where(_Str):
    pass


class RawWhere(Where):

    def __init__(self, query: str, *params: Data):
        self.query = query
        self.params = params

    def _str(self, params: Params):
        params.extend(self.params)
        return self.query


class WhereAnd(Where):

    def __init__(self, first: Where, second: Where):
        self.first = first
        self.second = second

    def _str(self, params: Params):
        first = self.first._str(params)
        second = self.second._str(params)
        return f'({first}) AND {second}'


class WhereEqual(RawWhere):
    '''
    Particular case of Where:
        key1 = value1 AND key2 = value2 AND ...
    Implemented separately for speed because it is very common 
    '''

    def __init__(self, **kwargs: Data):
        keys, values = unzip(kwargs)
        query = ' AND '.join(f'{k} = ?' for k in keys)
        super().__init__(query, *values)


class Everywhere(Where):

    def _str(self, params: Params):
        return ''


everywhere = Everywhere()  # Constant

EverywhereError = 'To prevent an accident, you must specify .where(..) for delete/update queries. If you really want everywhere use .everywhere()'


class QueryBody(_Str):

    _where: Optional[Where] = None
    _order_by: Tuple[str, ...] = ()
    _limit: Optional[int] = None

    def _str(self, params: List[Data]):
        body = []
        if self._where is not None:
            where_str = self._where._str(params)
            body.append(f'WHERE {where_str}')
        if self._order_by:
            by = ', '.join(self._order_by)
            body.append(f'ORDER BY {by}')
        if self._limit is not None:
            body.append('LIMIT ?')
            params.append(int(self._limit))
        return ' '.join(body)


F = TypeVar('F', bound=Callable)


def query_method(wrapped: F) -> F:
    '''
    Decorator for using Query methods from
    classes that inherit from it
    '''

    @wraps(wrapped)
    def wrapper(self, *args, **kwargs):
        if not hasattr(self, 'body'):
            self = Query(self, QueryBody())
        return wrapped(self, *args, **kwargs)

    return cast(F, wrapper)


K = TypeVar('K')
V = TypeVar('V')


def unzip(record: Dict[K, V]) -> Tuple[List[K], List[V]]:
    return map(list, zip(*record.items()))  # type:ignore


class Query:

    body: QueryBody

    def __init__(self, table: SqliteTable, body: QueryBody):
        self.table_name = table.name
        self.db = table.db
        self.body = body

    # "Where" related methods

    @query_method
    def where(self, **key_values):
        self.body._where = WhereEqual(**key_values)
        return self

    @query_method
    def raw_where(self, query: str, *params: Any):
        self.body._where = RawWhere(query, *params)
        return self

    @query_method
    def everywhere(self, **key_values):
        self.body._where = everywhere
        return self

    @query_method
    def order_by(self, *columns_asc_desc: str):
        self.body._order_by = columns_asc_desc
        return self

    @query_method
    def limit(self, limit: Optional[int]):
        self.body._limit = limit
        return self

    # Select method and mutation methods

    @query_method
    def select(self, *columns: str):
        what = ', '.join(columns) or '*'
        params = []
        where = self.body._str(params)
        query = (f'SELECT {what}' f' FROM {self.table_name} {where}')
        return self.db._execute(query, params)

    @query_method
    def delete(self):
        assert self.body._where is not None, EverywhereError
        params = []
        where = self.body._str(params)
        query = f'DELETE FROM {self.table_name} {where}'
        return self.db._execute(query, params)

    @query_method
    def insert(self, **record: Data):
        #inst = 'INSERT OR IGNORE' if ignore else 'INSERT'
        where = self.body._str([]).strip()
        assert not where, f'Unexpected statements for INSERT: {where}'
        table = self.table_name
        keys, values = unzip(record)
        columns = ', '.join(keys)
        marks = ', '.join('?' for _ in keys)
        query = f'INSERT INTO {table} ({columns}) VALUES ({marks})'
        return self.db._execute(query, values)

    @query_method
    def insert_or_ignore(self, **record: Data):
        where = self.body._str([]).strip()
        assert not where, f'Unexpected statements for INSERT: {where}'
        table = self.table_name
        keys, values = unzip(record)
        columns = ', '.join(keys)
        marks = ', '.join('?' for _ in keys)
        query = f'INSERT OR IGNORE INTO {table} ({columns}) VALUES ({marks})'
        cursor = self.db._execute(query, values)
        return cursor.rowcount

    @query_method
    def update_or_ignore(self, **partial_record: Data):
        assert self.body._where is not None, EverywhereError
        keys, params = unzip(partial_record)
        what = ', '.join(f'{c} = ?' for c in keys)
        where = self.body._str(params)
        query = f'UPDATE {self.table_name} SET {what} {where}'
        cursor = self.db._execute(query, params)
        return cursor.rowcount

    @query_method
    def update(self, **partial_record: Data):
        if not self.update_or_ignore(**partial_record):
            self.insert(**partial_record)
        return

    @query_method
    def count(self):
        return int(self.value('count(*)'))

    # high-level select methods

    @query_method
    def rows(self, *columns: str) -> List[DataRow]:
        return [*self.select(*columns)]

    @query_method
    def dicts(self, *columns: str) -> List[Record]:
        cursor = self.select(*columns)
        names = [c for c, *_ in cursor.description]
        return [dict(zip(names, row)) for row in cursor]

    @query_method
    def column(self, column: str) -> List[Data]:
        return [first for first, *_ in self.rows(column)]

    @query_method
    def value(self, column: str) -> Data:
        return self.limit(1).column(column)[0]

    @query_method
    def dict(self, *columns: str) -> Record:
        return self.limit(1).dicts(*columns)[0]

    @query_method
    def get_value(self, column: str) -> Optional[Data]:
        col = self.limit(1).column(column)
        return col[0] if col else None

    @query_method
    def get_dict(self, *columns: str) -> Optional[Record]:
        dicts = self.limit(1).dicts(*columns)
        return dicts[0] if dicts else None

    @query_method
    def series(self, *columns: str) -> Dict[str, List[Data]]:
        cursor = self.select(*columns)
        keys = [x[0] for x in cursor.description]
        out = {key: [] for key in keys}
        for row in cursor:
            for key, value in zip(keys, row):
                out[key].append(value)
        return out

    # random selection methods

    @query_method
    def random_rows(self, n: int, *columns: str):
        rows, _ = self._select_random(n, *columns)
        return rows

    @query_method
    def random_dicts(self, n: int, *columns: str) -> List[Record]:
        rows, keys = self._select_random(n, *columns)
        return [dict(zip(keys, row)) for row in rows]

    @query_method
    def random_dict(self, *columns: str):
        return self.random_dicts(1, *columns)[0]

    @query_method
    def _select_random(self, n: int, *columns: str):
        assert n > 0, 'Empty selection unsupported'
        N: int = self.count()
        assert N > 0, 'The table is empty'
        where = self.body._where
        column_names = []
        rows: List[List[Data]] = []
        while len(rows) < n:
            remaining = n - len(rows)
            # Filter many out randomly
            mod = 1 + N // (2 * remaining)
            print(mod)
            new_where = RawWhere(f'random() % ? = 0', mod)
            if where is not None:
                new_where = WhereAnd(where, new_where)
            self.body._order_by = ()
            self.body._where = new_where
            cursor = self.limit(remaining).select(*columns)
            column_names = [x[0] for x in cursor.description]
            rows.extend(cursor)
        shuffle(rows)
        return rows, column_names



from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from . import table_signatures
else:

    class IdentityProvider():

        def __getattribute__(self, _):
            return lambda f: f

    table_signatures = IdentityProvider()


from __future__ import annotations
import abc
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union
if TYPE_CHECKING:
    from .table import SqliteTable, Record, Data

#Columns = Optional[Excluding[str, Sequence[str]]] # If Excluding existed
Columns = Optional[Union[List[str], Tuple[str, ...]]]
OrderBy = Optional[Union[List[str], Tuple[str, ...]]]
Params = List[Any]


class WhereClause(abc.ABC):

    @abc.abstractmethod
    def parse_into(self, params: Params) -> str:
        ...

    @abc.abstractmethod
    def keys(self) -> List[str]:
        ...

    def __or__(self, other: WhereClause):
        return WhereCompound(self, 'OR', other)

    def __and__(self, other: WhereClause):
        return WhereCompound(self, 'AND', other)

    def __str__(self):
        params = []
        where_str = self.parse_into(params)
        return f'Where query string:\n{where_str}\nwith parameters:\n{params}'


class Where(WhereClause):
    cmp_operators = ['=', '<>', '<', '>=', '>', '<=', 'LIKE']

    def __init__(self, key: str, cmp: str, value: Any):
        assert cmp in self.cmp_operators
        self.key = key
        self.cmp = cmp
        self.value = value

    def parse_into(self, params: Params):
        params.append(self.value)
        return f'{self.key} {self.cmp} ?'

    def keys(self):
        return [self.key]


class WhereCompound(WhereClause):
    operators = ['AND', 'OR']

    def __init__(self, first: WhereClause, op: str, second: WhereClause):
        assert op in self.operators
        self.first = first
        self.op = op
        self.second = second

    def parse_into(self, params: Params):
        first = self.first.parse_into(params)
        second = self.second.parse_into(params)
        return f'({first}) {self.op} ({second})'

    def keys(self):
        return [*self.first.keys(), *self.second.keys()]


class WhereEQ(WhereClause):
    '''
    Particular case of Where and WhereCompound:
        
        key1 = value1 AND key2 = value2 AND ...

    Implemented separately for speed because it is very common 
    '''

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def parse_into(self, params: Params):
        keys, values = zip(*self.kwargs.items())
        params.extend(values)
        return ' AND '.join(f'{k} = ?' for k in keys)

    def keys(self):
        return [*self.kwargs.keys()]


class QueryPart(abc.ABC):

    @abc.abstractmethod
    def parse_into(self, params: Params) -> str:
        ...

    @abc.abstractmethod
    def keys(self) -> List[str]:
        ...

    def strict_check(self, table: SqliteTable, query_str: str):
        found = set(self.keys())
        expected = set(table.columns())
        diff = found.difference(expected)
        assert not diff, (f'Unknown columns for table {table.table_name}:\n'
                          f'{diff} not in {expected}\n'
                          f'Unsafe query (details above):\n{query_str}')
        return


class QueryBody(QueryPart):

    def __init__(self, complex_where: WhereClause = None,
                 order_by: OrderBy = None, limit: int = None, **where):
        if complex_where is None:
            _where = WhereEQ(**where) if where else None
        else:
            assert not where, ('Use one of (..., complex_where=...) or'
                               ' (..., **where), but not both')
            _where = complex_where
        self.order_by = order_by
        self.limit = limit
        self.where = _where

    def parse_into(self, params: Params):
        body = []
        if self.where is not None:
            where_str = self.where.parse_into(params)
            body.append(f'WHERE {where_str}')
        if self.order_by is not None:
            by = ', '.join(self.order_by)
            body.append(f'ORDER BY {by}')
        if self.limit is not None:
            body.append('LIMIT ?')
            params.append(int(self.limit))
        return ' '.join(body)

    def keys(self):
        keys = self.where.keys() if self.where else []
        if self.order_by is not None:
            for by in self.order_by:
                col, *s = by.split(maxsplit=1)
                keys.append(col)
                asc_desc = s[0].lower().strip()
                assert asc_desc in ('', 'asc', 'desc', 'ascending',
                                    'descending'), asc_desc
        return keys


class QueryHead(QueryPart):

    def keys(self):
        return []


class QueryHeadSelect(QueryHead):

    def __init__(self, columns: Columns, table: SqliteTable):
        self.table = table
        self.columns = columns

    def parse_into(self, params: Params):
        what = '*' if not self.columns else ', '.join(self.columns)
        return f'SELECT {what} FROM {self.table.table_name}'


class QueryHeadDelete(QueryHead):

    def __init__(self, table: SqliteTable):
        self.table = table

    def parse_into(self, params: Params):
        return f'DELETE FROM {self.table.table_name}'


class QueryHeadInsert(QueryHead):

    def __init__(self, record: Record, table: SqliteTable, ignore=False):
        self.table = table
        self.record = record
        self.inst = 'INSERT OR IGNORE' if ignore else 'INSERT'

    def parse_into(self, params: Params):
        table = self.table.table_name
        keys, values = zip(*self.record.items())
        columns = ', '.join(keys)
        marks = ', '.join('?' for _ in keys)
        params.extend(values)
        return f'{self.inst} INTO {table} ({columns}) VALUES ({marks})'

    def keys(self):
        return [*self.record.keys()]


class QueryHeadUpdate(QueryHead):

    def __init__(self, record: Record, table: SqliteTable):
        self.table = table
        self.record = record

    def parse_into(self, params: Params):
        table = self.table.table_name
        keys, values = zip(*self.record.items())
        what = ', '.join(f'{c} = ?' for c in keys)
        params.extend(values)
        return f'UPDATE {table} SET {what}'

    def keys(self):
        return [*self.record.keys()]


'''
Alternatives that I am considering for the future.
'''

# ---
'''
Number 3: BEST

table.where().order_by().limit().dicts()
table.where().order_by().limit().rows()
table.where().delete()


table.where(key=value1).order_by().limit().select()
table.where_str('blah = ?', param1, param2).order_by().limit().select()
'''

# ---
'''
Number 1: (discarded)
First the head then the body:

table.count()
table.count.where(...)() # awful

table.dicts[columns].where(key1=value1,...).order_by(key1, 'key2 desc',...).limit(3)()
table.rows[columns].where(key1=value1,...)
table.dict[columns].where_str('this() = ? and that=?', [3, 4])
table.update(key1=value1, ...).everywhere()
'''

# ---
'''
Number 2:
With a global variable (or module) Q, accessible also with table.Q
First the body, then the head (better):

table.count[where(), order_by()]()
table.value[where()], order_by()][limit[1](columns)
table.dicts[where(), order_by(), Q.limit(1)](columns)

Looks nice, but might be problematic for the type checker...

Example:

class Table:

    @property
    def dicts(self):
        return SelectColumns(self, post=to_dict)[limit(1)]

    @property
    def value(self):
        return SelectColumn(self, post=extractValue)[limit(1)]

    @property
    def delete(self):
        return MutateNoColumn(self)[WhereRequired]


class TableSelect:

    def __init__(self, table, post):
        self.table = table
        self.post = post
        self.where = None
        self.order_by = None
    
    def __getitem__(self, *args):
        for arg in args:
            if isinstance(arg, )
        check no overwrites
        check no overwrites
        self.where = ...
        self.order_by = ...
        return self


class SelectColumns(TableSelect):
    
    def __call__(*columns:str):
        data = ...
        return self.post(data)

class SelectColumn(TableSelect):
    
    def __call__(column:str):
        data = ...
        return self.post(data)
'''



from __future__ import annotations
import abc
from os import times
from typing import TYPE_CHECKING, Any, Dict, List, cast
from .database import SqliteDB, FilePath, custom_repr, Data
from .queries import (
    Columns,
    QueryBody,
    QueryHead,
    QueryHeadSelect,
    QueryHeadDelete,
    QueryHeadInsert,
    QueryHeadUpdate,
    Where,
)
from .table_signatures_ import table_signatures as T
# from ..methodtools import cached_property
Record = Dict[str, Data]


class SqliteTable:
    '''
    Sqlite table object.

    count(**kw) -> int
    columns() -> List[str]

    dicts(columns, **kw)      -> List[Record] : (Record=Dict[str, Data])
    column(column, **kw)      -> List[Data]
    series(columns, **kw)     -> Dict[str, List[Data]]
    rows(columns, **kw)       -> List[Tuple[Data]]
    
    dict(columns, **kw)       -> first Record (or Exception)
    get_dict(columns, **kw)   -> first Record | None

    value(column, **kw)       -> first value (or Exception)
    get_value(column, **kw)   -> first value | None
    '''

    def __init__(self, file: FilePath, table_name: str, strict: bool = False):
        self.file = file
        self.table_name = table_name
        self.db = SqliteDB(self.file)
        self.strict = strict

    def __repr__(self):
        return custom_repr(self, 'file', 'table_name')

    def __len__(self):
        return self.count()

    # Main high-level mutation methods

    def insert(self, partial_record: Record):
        body = QueryBody()
        head = QueryHeadInsert(partial_record, self)
        return self._run_query(head, body)

    def insert_or_ignore(self, partial_record: Record):
        body = QueryBody()
        head = QueryHeadInsert(partial_record, self, ignore=True)
        cursor = self._run_query(head, body)
        return cursor.rowcount

    @T.delete
    def delete(self, **kw):
        assert kw, 'Nowhere to del'
        body = QueryBody(**kw)
        head = QueryHeadDelete(self)
        cur = self._run_query(head, body)
        return cur.rowcount

    @T.update
    def update(self, partial_record: Record, **kw):
        if self.update_or_ignore(partial_record, **kw) == 0:
            self.insert(partial_record)
        return

    @T.update_or_ignore
    def update_or_ignore(self, partial_record: Record, **kw):
        body = QueryBody(**kw)
        head = QueryHeadUpdate(partial_record, self)
        cursor = self._run_query(head, body)
        return cursor.rowcount

    # Main high-level select methods

    def columns(self) -> List[str]:
        query_str = 'SELECT name FROM PRAGMA_TABLE_INFO(?)'
        return self.db.execute_column(query_str, [self.table_name])

    @T.dicts
    def dicts(self, *columns: str, **kw):
        body = QueryBody(**kw)
        return self._select_dicts(columns, body)

    @T.column
    def column(self, column: str, **kw):
        body = QueryBody(**kw)
        return self._select_column(column, body)

    @T.series
    def series(self, *columns: str, **kw):
        body = QueryBody(**kw)
        return self._select_series(columns, body)

    @T.rows
    def rows(self, *columns: str, **kw):
        body = QueryBody(**kw)
        return self._select_rows(columns, body)

    @T.dict
    def dict(self, *columns: str, **kw):
        return self.dicts(*columns, **kw)[0]

    @T.get_dict
    def get_dict(self, *columns: str, **kw):
        dicts = self.dicts(*columns, **kw)
        return dicts[0] if dicts else None

    @T.value
    def value(self, column: str, **kw):
        kw['limit'] = 1
        return self.column(column, **kw)[0]

    @T.get_value
    def get_value(self, column: str, **kw):
        values = self.column(column, limit=1, **kw)
        return values[0] if values else None

    @T.count
    def count(self, **kw):
        return int(self.value('count(*)'))

    # Random high-level select methods

    @T.random_dicts
    def random_dicts(self, *columns: str, limit: int, **kw):
        rows = self.random_rows(*columns, limit=limit, **kw)
        to_dict = self._to_dict_map(columns)
        return [to_dict(row) for row in rows]

    @T.random_dict
    def random_dict(self, *columns: str, **kw):
        return self.random_dicts(*columns, limit=1, **kw)[0]

    @T.random_rows
    def random_rows(self, *columns: str, limit: int, **kw):
        'Repetitions may occur but with low probability'
        body = QueryBody(**kw)
        return self._select_random_rows(columns, int(limit), body)

    # Low-level select methods

    def _run_query(self, head: QueryHead, body: QueryBody):
        params: List[Any] = []
        head_str = head.parse_into(params)
        body_str = body.parse_into(params)
        query_str = f'{head_str} {body_str}'
        if self.strict:
            head.strict_check(self, query_str)
            body.strict_check(self, query_str)
        return self.db._execute(query_str, params)

    def _select_rows(self, columns: Columns, body: QueryBody):
        head = QueryHeadSelect(columns, self)
        return [*self._run_query(head, body)]

    def _select_dicts(self, columns: Columns, body: QueryBody):
        rows = self._select_rows(columns, body)
        to_dict = self._to_dict_map(columns)
        return [to_dict(row) for row in rows]

    def _to_dict_map(self, columns: Columns):
        columns = columns or self.columns()
        to_dict = lambda row: dict(zip(columns, row))
        return to_dict

    def _select_column(self, column: str, body: QueryBody):
        rows = self._select_rows([column], body)
        return [first for first, *_ in rows]

    def _select_series(self, columns: Columns, body: QueryBody):
        keys = self.columns() if columns is None else columns
        out = {key: [] for key in keys}
        for row in self._select_rows(columns, body):
            for key, value in zip(keys, row):
                out[key].append(value)
        return out

    def _select_random_rows(self, columns: Columns, limit: int,
                            body: QueryBody):
        n_total: int = self._select_column('count(*)', body)[0]
        assert n_total > 0 or limit == 0, f'The table is empty'
        head = QueryHeadSelect(columns, self)
        body.order_by = ['random()']
        initial_where = body.where
        rows: List[List[Data]] = []
        while len(rows) < limit:
            assert limit is not None
            if n_total - limit < 1000:
                body.limit = limit
                body.where = initial_where
                more = [*self._run_query(head, body)]
            else:
                div = 1 + n_total // (2 * limit)
                new_where = Where(f'random() % {div}', '=', 0)
                body.limit = limit
                body.where = new_where if not initial_where else initial_where & new_where
                more = [*self._run_query(head, body)]
            rows.extend(more)
        return rows


def test():
    import random
    table = SqliteTable('.test.db', 'rand_table')
    table.db.execute(f'''
        CREATE TABLE IF NOT EXISTS rand_table(
            key int NOT NULL PRIMARY KEY,
            age int NOT NULL,
            prob double NOT NULL
        )
    ''')

    print(len(table))
    for k in set(range(1000)).difference(set(table.column('key'))):
        record = dict(
            key=k,
            age=random.randint(0, 5),
            prob=round(random.random(), 3),
        )
        table.insert(record)
    print(table.random_dicts(limit=3))
    print(table.random_dicts(limit=3))
    print(len(table))


import sqlite3
from typing import TYPE_CHECKING, Generic, TypeVar, Dict, List, Optional, cast
from .table import SqliteTable, Record, Columns, Data
from .queries import WhereClause, OrderBy
'''
Type declaration for several methods in table.py
Just replace from it
    **kw
with
    complex_where: WhereClause = None,
    order_by: OrderBy = None,
    limit: int = None,
    **where: Data
except for:
    value, dict, get_dict do not admit limit (forced to 1)
    random require non null limit
'''

F = TypeVar('F')


class copy(Generic[F]):

    def __init__(self, _: F) -> None:
        ...

    def __call__(self, wrapped) -> F:
        return wrapped


def _dicts(
    self: SqliteTable,
    *columns: str,
    complex_where: WhereClause = None,
    order_by: OrderBy = None,
    limit: int = None,
    **where: Data,
) -> List[Record]:
    ...


dicts = copy(_dicts)


def _column(
    self: SqliteTable,
    column: str,
    complex_where: WhereClause = None,
    order_by: OrderBy = None,
    limit: int = None,
    **where: Data,
) -> List[Data]:
    ...


column = copy(_column)


def _series(
    self: SqliteTable,
    *columns: str,
    complex_where: WhereClause = None,
    order_by: OrderBy = None,
    limit: int = None,
    **where: Data,
) -> Dict[str, List[Data]]:
    ...


series = copy(_series)


def _rows(
    self: SqliteTable,
    *columns: str,
    complex_where: WhereClause = None,
    order_by: OrderBy = None,
    limit: int = None,
    **where: Data,
) -> List[List[Data]]:
    ...


rows = copy(_rows)


def _dict(
    self: SqliteTable,
    *columns: str,
    complex_where: WhereClause = None,
    **where: Data,
) -> Record:
    ...


dict = copy(_dict)


def _get_dict(
    self: SqliteTable,
    *columns: str,
    complex_where: WhereClause = None,
    **where: Data,
) -> Optional[Record]:
    ...


get_dict = copy(_get_dict)


def _value(
    self: SqliteTable,
    column: str,
    complex_where: WhereClause = None,
    **where: Data,
) -> Data:
    ...


value = copy(_value)


def _get_value(
    self: SqliteTable,
    column: str,
    complex_where: WhereClause = None,
    **where: Data,
) -> Optional[Data]:
    ...


get_value = copy(_get_value)


def _count(
    self: SqliteTable,
    complex_where: WhereClause = None,
    limit: int = None,
    **where: Data,
) -> int:
    ...


count = copy(_count)


def _delete(
    self: SqliteTable,
    complex_where: WhereClause = None,
    limit: int = None,
    **where: Data,
) -> int:
    ...


delete = copy(_delete)


def _update_or_ignore(
    self: SqliteTable,
    partial_record: Record,
    complex_where: WhereClause = None,
    order_by: OrderBy = None,
    limit: int = None,
    **where: Data,
) -> int:
    ...


update_or_ignore = copy(_update_or_ignore)


def _update(
    self: SqliteTable,
    partial_record: Record,
    complex_where: WhereClause = None,
    order_by: OrderBy = None,
    limit: int = None,
    **where: Data,
) -> None:
    ...


update = copy(_update)


def _random_dict(
    self: SqliteTable,
    *columns: str,
    complex_where: WhereClause = None,
    **where: Data,
) -> Record:
    ...


random_dict = copy(_random_dict)


def _random_dicts(
    self: SqliteTable,
    *columns: str,
    limit: int,
    complex_where: WhereClause = None,
    **where: Data,
) -> List[Record]:
    ...


random_dicts = copy(_random_dicts)


def _random_rows(
    self: SqliteTable,
    *columns: str,
    limit: int,
    complex_where: WhereClause = None,
    **where: Data,
) -> List[List[Data]]:
    ...


random_rows = copy(_random_rows)



from __future__ import annotations
from typing import List, Optional, Tuple
from contextlib import contextmanager
import json, time, random, logging
from .table import Record, SqliteTable, FilePath, Data


def create_deadline(timeout: Optional[float]):
    if timeout is None:
        return float('inf')
    return time.time() + timeout


class TimeoutError(Exception):
    message = 'Waiting for access token timed out'


class TokenError(Exception):
    message = 'Invalid or expired token'


class SqliteStore(SqliteTable):

    def __init__(self, file: FilePath, name: str):
        super().__init__(file, name)
        self.db.execute(f'''
        CREATE TABLE IF NOT EXISTS {self.name}(
            key text NOT NULL PRIMARY KEY,
            value text,
            lock_token double NOT NULL,
            locked_until double NOT NULL
        )
        ''')
        the_columns = {'key', 'value', 'lock_token', 'locked_until'}
        assert set(self.columns()) == the_columns, self.columns()
        return

    def __getitem__(self, key: str):
        d = super().where(key=key).get_dict('value')
        if not d:
            raise KeyError(key)
        return json.loads(d['value'] or 'null')

    def get(self, key: str, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def values(self):
        return [json.loads(s or 'null') for s in super().column('value')]

    def keys(self) -> List[str]:
        return [k for k in super().column('key')]

    def items(self) -> List[Tuple[str, Data]]:
        items = super().rows('key', 'value')
        return [(k, json.loads(v or 'null')) for k, v in items]

    def __set_assuming_token(self, key: str, value: Data):
        return super().where(key=key).update(value=json.dumps(value))

    def __del_assuming_token(self, key: str):
        return SqliteTable.delete(self.where(key=key))

    def _current_lock(self, key: str):
        return super().where(key=key).get_dict('lock_token', 'locked_until')

    def set(self, key: str, value: Data, token: float):
        '''
        Requires a token provided by the exclusive access context
        manager self.wait_token() or self.ask_token().
        '''
        with self.assert_token(key, token=token):
            self.__set_assuming_token(key, value)
        return

    def delete(self, key: str, token: float):
        '''
        Requires a token provided by the exclusive access context
        manager self.wait_token() or self.ask_token().
        '''
        with self.assert_token(key, token=token):
            self.__del_assuming_token(key)
        return

    def ask_del(self, key: str):
        with self.ask_token(key) as token:
            if not token:
                return False
            self.__del_assuming_token(key)
        return False

    def ask_set(self, key: str, value: Data):
        with self.ask_token(key) as token:
            if not token:
                return False
            self.__set_assuming_token(key, value)
        return False

    def wait_del(self, key: str, request_every: float = 0.02,
                 timeout: Optional[float] = 3):
        wait = self.wait_token(key, request_every=request_every,
                               timeout=timeout)
        with wait:
            self.__del_assuming_token(key)

    def wait_set(self, key: str, value: Data, request_every: float = 0.02,
                 timeout: Optional[float] = 3):
        wait = self.wait_token(key, request_every=request_every,
                               timeout=timeout)
        with wait as token:
            self.set(key, value, token)

    @contextmanager
    def ask_token(self, *keys: str, max_duration: float = 0.5):
        '''
        with table.ask_token('some_key') as token:
            if not token:
                # The resource is locked by other
            else:
                # The resource is mine
                table.set('some_key', some_value, token)
        '''
        assert keys, 'You must specify keys to be locked explicitely'
        token = random.random()
        try:
            gained_access = all([
                self._ask_access(key, token=token, max_duration=max_duration)
                for key in keys
            ])
            yield token if gained_access else None
        finally:
            for key in keys:
                self._unlock(key, token, max_duration)
        return

    @contextmanager
    def assert_token(self, *keys: str, token: float):
        assert keys, 'You must specify keys to be locked explicitely'
        max_duration = 0.5
        try:
            for key in keys:
                if not self._ask_access(key, token=token,
                                        max_duration=max_duration):
                    raise TokenError(key, token)
            yield
        finally:
            for key in keys:
                self._unlock(key, token, max_duration)
        return

    @contextmanager
    def wait_token(self, *keys: str, max_duration: float = 0.5,
                   request_every: float = 0.02, timeout: Optional[float] = 3,
                   _warn_del=True):
        '''
        # Wait at most {timeout} seconds to get exclusive access,
        # requesting every {request_every} seconds
        with table.wait_token('some_key') as token:
            current = table['some_key']
            ...
            # No one else can set some_key
            # during next {max_duration} seconds
            ...
            current = table.set('some_key', some_value, token)
        '''
        assert keys, 'You must specify keys to be locked explicitely'
        token = random.random()
        try:
            for key in keys:
                deadline = create_deadline(timeout)
                while not self._ask_access(key, token, max_duration):
                    if time.time() > deadline:
                        raise TimeoutError(timeout)
                    time.sleep(request_every)
            yield token
        finally:
            for key in keys:
                self._unlock(key, token, max_duration)
        return

    def _ask_access(self, key: str, token: float, max_duration: float):
        now = time.time()
        until = now + max_duration
        # Compete with other processes for an exclusive update
        cursor = self.db._execute(
            f'''
        UPDATE {self.name} SET
            lock_token=?,
            locked_until=?
        WHERE
            key=? AND (
                lock_token<0 OR
                lock_token=? OR
                locked_until<?
            )
        ''', [token, until, key, token, now])
        if cursor.rowcount > 0:  # Race winner
            return True
        # Maybe the key was not even present:
        return self.insert_or_ignore(
            key=key,
            lock_token=token,
            locked_until=until,
        )

    def _unlock(self, key: str, token: float, max_duration: float):
        d = self._current_lock(key)
        if not d:
            # Entry was deleted (and unlocked)
            return
        remaining = d['locked_until'] - time.time()
        if remaining < 0:
            ms = round(-remaining * 1000)
            logging.warning(
                f'Locked {repr(key)} during {ms}ms more than max_duration={max_duration}s'
            )
        if d['lock_token'] == token:
            super().where(key=key).update(lock_token=-1)
        return


def test():
    store = SqliteStore('.test.db', 'the_table')
    print(store.db.table_names())
    print(len(store))
    store.wait_set('carlos', 'HELLO')
    print(store.dicts())
    with store.wait_token('carlos', 'adri') as token:
        print(f'I HAVE ACCESS! {token}')
        with store.ask_token('carlos') as access:
            assert access is None

    store.wait_set('carlos', 'HELLO2')
    store.wait_set('adri', 'BYE')
    store.wait_set('santiago', 'HMM')
    key = store.random_rows(1, 'key')[0][0]
    print(f'Deleting key={key}')
    store.delete(key, token)
    print('OK')
